"""
Requirements:
pip install wandb
export WANDB_API_KEY=XXX
"""

import pdb

import wandb
import numpy as np
import matplotlib
matplotlib.use('agg')
from matplotlib import pyplot as plt
import argparse

from sys import argv
import os

plt.style.use('ggplot')


parser = argparse.ArgumentParser(description="generate plots")
parser.add_argument("--min-obs-delay", type=int, default=0)
parser.add_argument("--max-obs-delay", type=int, default=1)
parser.add_argument("--min-act-delay", type=int, default=0)
parser.add_argument("--max-act-delay", type=int, default=1)
parser.add_argument("--gc", type=float, default=266)

config = parser.parse_args()



def flatten_dict(d, sep='.'):
  """specifically for mongodb queries"""
  r = {}
  for k, v in d.items():
    if isinstance(v, dict) and all(not l.startswith('$') for l in v):
      r.update((k + sep + l, w) for l, w in flatten_dict(v).items())
    else:
      r[k] = v
  return r


class Figure:
  def __init__(self, size=None, dpi=None):
    self.dpi = dpi
    self.size = size
    f, ax = plt.subplots(figsize=self.size, dpi=self.dpi)
    self.fig: plt.Figure = f
    self.ax: plt.Axes = ax

  def rgb(self, height=None, width=None):
    fig = self.fig
    from matplotlib.backends.backend_agg import FigureCanvasAgg
    fig.canvas.switch_backends(FigureCanvasAgg)
    w, h = fig.canvas.get_width_height()
    w, h = width or w, height or h
    fig.set_size_inches(w / fig.get_dpi(), h / fig.get_dpi())
    fig.canvas.draw()
    buf = fig.canvas.tostring_rgb()  # alternatively tostring_argb
    rgb = np.fromstring(buf, np.uint8).reshape(h, w, 3)
    # plt.close(fig)
    return rgb

  def save(self, path, **kwargs):
    self.fig.savefig(path, **kwargs)
    # plt.close(self.fig)

  def __del__(self):
    plt.close(self.fig)


def smooth(a, window=0.1):
  n = int(len(a) * window + 1)
  padded_sequence = np.concatenate(([a[0]]*(n//2), a, [a[-1]]*(n//2)))
  return np.convolve(np.ones(n)/n, padded_sequence, mode='valid')

def plot_percentile(ys):
  x = range(ys.shape[1])
  f = Figure()
  f.ax.plot(x, np.nanmedian(ys, axis=0))
  ya = np.nanpercentile(ys, 25, axis=0)
  yb = np.nanpercentile(ys, 75, axis=0)
  f.ax.fill_between(x, ya, yb, alpha=0.25)
  return f

def mean_ci90(data, n):
  tval = [6.314, 2.920, 2.353, 2.132, 2.015, 1.943, 1.895]
  f = tval[n-1] / np.sqrt(n)
  ci = np.nanstd(data, axis=0) * f
  return ci

def plot_mean_half_std(f, ys, label=None, xmul=1):
  print(f"DEBUG: label:{label}")
  if label == "SAC_naive":
    col = 'lightpink'
  elif label == "SAC":
    col = 'orange'
  elif label == "SAC_undelayed":
    col = 'deepskyblue'
  elif label == "DC/AC":
    col = 'limegreen'
  elif label == "RTAC":
    col = 'silver'
  ys = np.asarray([smooth(y, 0.1) for y in ys])
  m = np.nanmean(ys, axis=0)
  x = np.arange(len(m)) * xmul
  f.ax.plot(x, m, label=label, color=col)
  std = 0.5 * np.nanstd(ys, axis=0)
  f.ax.fill_between(x, m-std, m+std, alpha=0.25, color=col)

def plot_mean_ci90(f, ys, label=None, xmul=1):
  print(f"DEBUG: label:{label}")
  #if label == "SAC_naive":
  #  col = 'lightpink'
  if label == "SAC":
    col = 'deepskyblue'
  elif label == "SAC_undelayed":
    col = 'lightpink'
  elif label == "DC/AC":
    col = 'limegreen'
  elif label == "RTAC":
    col = 'orange'
  else:
    raise ValueError("unknown label")
  ys = np.asarray([smooth(y, 0.1) for y in ys])
  m = np.nanmean(ys, axis=0)
  x = np.arange(len(m)) * xmul
  f.ax.plot(x, m, label=label, color=col)
  ci = mean_ci90(ys, len(ys))
  f.ax.fill_between(x, m-ci, m+ci, alpha=0.25, color=col)

def plot_mean_conf95(f, ys, label=None, xmul=1):
  """https://en.wikipedia.org/wiki/Confidence_interval#Basic_steps"""
  ys = np.asarray([smooth(y, 0.1) for y in ys])
  m = np.nanmean(ys, axis=0)
  x = np.arange(len(m)) * xmul
  f.ax.plot(x, m, label=label)
  conf = 1.96 * np.nanstd(ys, axis=0) / np.sqrt(len(ys))
  f.ax.fill_between(x, m-conf, m+conf, alpha=0.25)

api = wandb.Api()

def preproc_return(r):
  xprev = 0.
  for x in r:
    if np.isnan(x) or x == 'NaN':
      x = xprev
    yield x
    xprev = x

def get_returns(entity='drl', proj='delay', **kwargs):
  query = [{"config." + k.replace("._", ".+"): v} for k, v in flatten_dict(kwargs, sep='.').items()]

  mongo_db_query = {"$and": [
    {"$or": [
      {"state": "finished"},
      {"state": "running"},
    ]},
  *query]}

  # print("Query:", mongo_db_query)
  runs = api.runs(entity + "/" + proj, mongo_db_query)
  # for r in runs:
      # print(r.id)

  if not runs:
    return []
  returns = [r.history(samples=10000).returns_test for r in runs]
  min_len = min(len(r) for r in returns)
  returns = [r[:min_len] for r in returns]
  returns = [list(preproc_return(r)) for r in returns]
  assert all([isinstance(x, float) for r in returns for x in r]), f"{returns}"
  return np.array(returns)


def compare():
  envs = ("Humanoid-v2", "Ant-v2", "HalfCheetah-v2", "Walker2d-v2", "Hopper-v2", "Reacher-v2")
  sub_envs = ("Ant-v2", "HalfCheetah-v2", "Walker2d-v2")
  # 0303 2323 (0, 1, 0, 1), (0, 2, 0, 1), (0, 1, 0, 2), (1, 2, 1, 2), (0, 3, 0, 3)
  gc = config.gc  # git commit count
  for env_id in envs:
    env_dict = dict(id=env_id, frame_skip=0, min_action_delay=config.min_act_delay, sup_action_delay=config.max_act_delay, min_observation_delay=config.min_obs_delay, sup_observation_delay=config.max_obs_delay)
    rets = {}
    if env_id in sub_envs:
        rets["SAC"] = get_returns(git=dict(count={"$gte": gc}), Agent=dict(_="agents.sac:Agent", Model=dict(tbmdp=None)), Env=env_dict)
        #rets["SAC_naive"] = get_returns(git=dict(count={"$gte": gc}), Agent=dict(_="agents.sac:Agent", Model=dict(tbmdp=True)), Env=env_dict)
    else:
        rets["SAC"] = get_returns(git=dict(count={"$gte": gc}), Agent=dict(_="agents.sac:Agent"), Env=env_dict)
    rets["DC/AC"] = get_returns(git=dict(count={"$gte": gc}), Agent=dict(_="agents.drtac:Agent", rtac=False), Env=env_dict)
    rets["RTAC"] = get_returns(git=dict(count={"$gte": gc}), Agent=dict(_="agents.drtac:Agent", rtac=True), Env=env_dict)
    #rets["SAC_undelayed"] = get_returns(git=dict(count={"$gte": gc}), Agent=dict(_="agents.sac_undelayed:Agent"), Env=env_dict)


    print('For', env_id, 'found\n' + ',\n'.join(f'  {len(v)} experiments for {k}' for k, v in rets.items()) + '.')

    # f = Figure((6, 4), 300)
    f = Figure((4, 3), 300)
    [plot_mean_ci90(f, r, xmul=0.002, label=lab) for lab, r in rets.items()]

    if env_id == "Reacher-v2":
      # f.ax.legend(("JSAC", "SAC", "SAC (no delay)", "SAC (no delay, normalized)"))
      f.ax.legend(rets.keys(), facecolor="white", loc="lower right")
    plt.title(env_id)
    plt.grid(color="white")
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    plt.ylabel('average return')
    plt.xlabel('million steps')
    plt.tight_layout()
    f.save(env_id + ".png")


compare()

"""
To concatenate the plots run
convert Ant-v2.png HalfCheetah-v2.png Walker2d-v2.png +append s.png && convert Hopper-v2.png Humanoid-v2.png Reacher-v2.png +append t.png && convert s.png t.png -append st.png; rm s.png t.png
"""
